# encoding: UTF-8
import operator


def check_regression_fixture_workflow(
    testdir,
    source,
    data_getter,
    data_modifier,
    expected_data_1,
    expected_data_2,
    compare_fn=None,
):
    """
    Helper method to test regression fixtures like `data_regression`. Offers a basic template/script
    able to validate main behaviors expected by regression fixtures.

    Usage
    -----

    ```
    import sys

    monkeypatch.setattr(sys, 'get_data', lambda: 'foo', raising=False)
    source = '''
        import sys
        def test_1(fake_regression):
            data = sys.get_data()
            fake_regression.Check(data)
    '''

    def get_data():
        fake_filename = os.path.join(str(testdir.tmpdir), 'test_file', 'test_1.fake')
        assert os.path.isfile(fake_filename)
        with open(fake_filename) as f:
            return f.read()

    check_regression_fixture_workflow(
        testdir,
        source,
        data_getter=get_data,
        data_modifier=lambda: monkeypatch.setattr(sys, 'get_data', lambda: 'bar', raising=False),
        expected_data_1='foo',
        expected_data_2='bar',
    )
    ```

    :param Testdir testdir: `testdir` fixture. Requires pytest's `pytester` to be installed.
    :param str source: Source code using regression fixture.
    :param callable data_getter: Function without arguments that returns contents of file
        created by regression test when it fails first time (i.e. the expected file for future
        runs).
    :param callable data_modifier: Function without arguments that must change data compared by
        regression fixture so it fails in next comparison.
    :param object expected_data_1: Expected data in expected file for first state of data.
    :param object expected_data_2: Expected data in expected file for second state of data.
    :param callable compare_fn: function with signature (obtained, expected) used to ensure
        both data are equal. By default uses operator.eq, but can customized (for example
        to compare numpy arrays).
    """
    if compare_fn is None:
        compare_fn = operator.eq
    testdir.makepyfile(test_file=source)

    # First run fails because there's no expected file yet
    result = testdir.inline_run()
    result.assertoutcome(failed=1)

    # ensure now that the file was generated and the test passes
    xx = data_getter()
    compare_fn(xx, expected_data_1)
    result = testdir.inline_run()
    result.assertoutcome(passed=1)

    # changing the regression data makes the test fail (file remains unchanged)
    data_modifier()
    result = testdir.inline_run()
    result.assertoutcome(failed=1)
    compare_fn(data_getter(), expected_data_1)

    # force regeneration (test fails again)
    result = testdir.inline_run("--force-regen")
    result.assertoutcome(failed=1)
    compare_fn(data_getter(), expected_data_2)

    # test should pass again
    result = testdir.inline_run()
    result.assertoutcome(passed=1)



import io
from functools import partial
from pytest_regressions.common import Path

from pytest_regressions.testing import check_regression_fixture_workflow


def test_image_regression(image_regression, datadir):
    import matplotlib

    # this ensures matplot lib does not use a GUI backend (such as Tk)
    matplotlib.use("Agg")

    import matplotlib.pyplot as plt
    import numpy as np

    t = np.arange(0.0, 2.0, 0.01)
    s = 1 + np.sin(2 * np.pi * t)

    fig, ax = plt.subplots()
    ax.plot(t, s)

    ax.set(
        xlabel="time (s)",
        ylabel="voltage (mV)",
        title="About as simple as it gets, folks",
    )
    ax.grid()

    image_filename = datadir / "test.png"
    fig.savefig(str(image_filename))

    image_regression.check(image_filename.read_bytes(), diff_threshold=1.0)


def test_image_regression_workflow(testdir, monkeypatch, datadir):
    """
    :type testdir: _pytest.pytester.TmpTestdir
    :type monkeypatch: _pytest.monkeypatch.monkeypatch
    """
    import sys
    from PIL import Image

    def get_image(color):
        f = io.BytesIO()
        img = Image.new("RGB", (100, 100), color)
        img.save(f, "PNG")
        return f.getvalue()

    monkeypatch.setattr(sys, "get_image", partial(get_image, "white"), raising=False)
    source = """
        import sys
        def test_1(image_regression):
            contents = sys.get_image()
            image_regression.check(contents)
    """

    def get_file_contents():
        fn = Path(str(testdir.tmpdir)) / "test_file" / "test_1.png"
        assert fn.is_file()
        return fn.read_bytes()

    check_regression_fixture_workflow(
        testdir,
        source,
        data_getter=get_file_contents,
        data_modifier=lambda: monkeypatch.setattr(
            sys, "get_image", partial(get_image, "black"), raising=False
        ),
        expected_data_1=partial(get_image, "white"),
        expected_data_2=partial(get_image, "black"),
    )



# encoding: UTF-8

import six

from pytest_regressions.common import perform_regression_check


class NumericRegressionFixture(object):
    """
    Numeric Data Regression fixture implementation used on num_regression fixture.
    """

    DISPLAY_PRECISION = 17  # Decimal places
    DISPLAY_WIDTH = 1000  # Max. Chars on outputs

    def __init__(self, datadir, original_datadir, request):
        """
        :type datadir: Path
        :type original_datadir: Path
        :type request: FixtureRequest
        """
        self._tolerances_dict = {}
        self._default_tolerance = {}

        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self._force_regen = False

        self._pandas_display_options = (
            "display.precision",
            NumericRegressionFixture.DISPLAY_PRECISION,
            "display.width",
            NumericRegressionFixture.DISPLAY_WIDTH,
        )

    def _check_data_types(self, key, obtained_column, expected_column):
        """
        Check if data type of obtained and expected columns are the same. Fail if not.
        Helper method used in _check_fn method.
        """
        import numpy as np

        __tracebackhide__ = False

        obtained_data_type = obtained_column.values.dtype
        expected_data_type = expected_column.values.dtype
        if obtained_data_type != expected_data_type:
            # Check if both data types are comparable as numbers (float, int, short, bytes, etc...)
            if np.issubdtype(obtained_data_type, np.number) and np.issubdtype(
                expected_data_type, np.number
            ):
                return

            # In case they are not, assume they are not comparable
            error_msg = (
                "Data type for data %s of obtained and expected are not the same.\n"
                "Obtained: %s\n"
                "Expected: %s\n" % (key, obtained_data_type, expected_data_type)
            )
            raise AssertionError(error_msg)

    def _check_data_shapes(self, obtained_column, expected_column):
        """
        Check if obtained and expected columns have the same size.
        Helper method used in _check_fn method.
        """
        __tracebackhide__ = False

        obtained_data_shape = obtained_column.values.shape
        expected_data_shape = expected_column.values.shape
        if obtained_data_shape != expected_data_shape:
            error_msg = (
                "Obtained and expected data shape are not the same.\n"
                "Obtained: %s\n"
                "Expected: %s\n" % (obtained_data_shape, expected_data_shape)
            )
            raise AssertionError(error_msg)

    def _check_fn(self, obtained_filename, expected_filename):
        """
        Check if dict contents dumped to a file match the contents in expected file.

        :param six.text_type obtained_filename:
        :param six.text_type expected_filename:
        """
        import numpy as np
        import pandas as pd

        __tracebackhide__ = False

        obtained_data = pd.read_csv(six.text_type(obtained_filename))
        expected_data = pd.read_csv(six.text_type(expected_filename))

        comparison_tables_dict = {}
        for k in obtained_data.keys():
            obtained_column = obtained_data[k]
            expected_column = expected_data.get(k)

            if expected_column is None:
                error_msg = "Could not find key '%s' in the expected results.\n" % (k,)
                error_msg += "Keys in the obtained data table: ["
                for k in obtained_data.keys():
                    error_msg += "'%s', " % (k,)
                error_msg += "]\n"
                error_msg += "Keys in the expected data table: ["
                for k in expected_data.keys():
                    error_msg += "'%s', " % (k,)
                error_msg += "]\n"
                error_msg += "To update values, use --force-regen option.\n\n"
                raise AssertionError(error_msg)

            tolerance_args = self._tolerances_dict.get(k, self._default_tolerance)

            self._check_data_types(k, obtained_column, expected_column)
            self._check_data_shapes(obtained_column, expected_column)

            data_type = obtained_column.values.dtype
            if data_type in [float, np.float, np.float16, np.float32, np.float64]:
                not_close_mask = ~np.isclose(
                    obtained_column.values,
                    expected_column.values,
                    equal_nan=True,
                    **tolerance_args
                )
            else:
                not_close_mask = obtained_column.values != expected_column.values

            if np.any(not_close_mask):
                diff_ids = np.where(not_close_mask)[0]
                diff_obtained_data = obtained_column[diff_ids]
                diff_expected_data = expected_column[diff_ids]
                diffs = np.abs(obtained_column - expected_column)[diff_ids]

                comparison_table = pd.concat(
                    [diff_obtained_data, diff_expected_data, diffs], axis=1
                )
                comparison_table.columns = [
                    "obtained_%s" % (k,),
                    "expected_%s" % (k,),
                    "diff",
                ]
                comparison_tables_dict[k] = comparison_table

        if len(comparison_tables_dict) > 0:
            error_msg = "Values are not sufficiently close.\n"
            error_msg += "To update values, use --force-regen option.\n\n"
            for k, comparison_table in comparison_tables_dict.items():
                error_msg += "%s:\n%s\n\n" % (k, comparison_table)
            raise AssertionError(error_msg)

    def _dump_fn(self, data_object, filename):
        """
        Dump dict contents to the given filename

        :param pd.DataFrame data_object:
        :param six.text_type filename:
        """
        data_object.to_csv(
            six.text_type(filename),
            float_format="%%.%sg" % (NumericRegressionFixture.DISPLAY_PRECISION,),
        )

    def check(
        self,
        data_dict,
        basename=None,
        fullpath=None,
        tolerances=None,
        default_tolerance=None,
        data_index=None,
        fill_different_shape_with_nan=True,
    ):
        """
        Checks the given dict against a previously recorded version, or generate a new file.
        The dict must map from user-defined keys to 1d numpy arrays.

        Example:
        num_regression.check({
            'U_gas': U[0][positions],
            'U_liquid': U[1][positions],
            'gas_vol_frac [-]': vol_frac[0][positions],
            'liquid_vol_frac [-]': vol_frac[1][positions],
            'P': Pa_to_bar(P)[positions],
        })

        :param dict data_dict: dict mapping keys to numpy arrays.
        :param six.text_type basename: basename of the file to test/record. If not given the name
            of the test is used.
        :param six.text_type fullpath: complete path to use as a reference file. This option
            will ignore embed_data completely, being useful if a reference file is located
            in the session data dir for example.
        :param dict tolerances: dict mapping keys from the data_dict to tolerance settings for the
            given data. Example: tolerances={'U': Tolerance(atol=1e-2)}
        :param dict default_tolerance: dict mapping the default tolerance for the current check
            call. Example: default_tolerance=dict(atol=1e-7, rtol=1e-18).
            If not provided, will use defaults from numpy's isclose function.
        :param list data_index: If set, will override the indexes shown in the outputs. Default
            is panda's default, which is range(0, len(data))
        :param bool fill_different_shape_with_nan: If set, all the data provided in the data_dict
            that has size lower than the bigger size will be filled with np.NaN, in order to save
            the data in a CSV file.

        basename and fullpath are exclusive.
        """
        import numpy as np
        import pandas as pd
        import functools

        __tracebackhide__ = False

        if tolerances is None:
            tolerances = {}
        self._tolerances_dict = tolerances

        if default_tolerance is None:
            default_tolerance = {}
        self._default_tolerance = default_tolerance

        data_shapes = []
        for obj in data_dict.values():
            assert type(obj) in [
                np.ndarray
            ], "Only numpy arrays are valid for numeric_data_regression fixture.\n"
            shape = obj.shape

            assert len(shape) == 1, (
                "Only 1D arrays are supported on num_data_regression fixture.\n"
                "Array with shape %s was given.\n" % (shape,)
            )
            data_shapes.append(shape[0])

        if not np.all(data_shapes == data_shapes[0]):
            if not fill_different_shape_with_nan:
                assert (
                    False
                ), "Data dict with different array lengths will not be accepted." "Try setting fill_different_shape_with_nan=True."
            else:
                max_size = max(data_shapes)
                for k, obj in data_dict.items():
                    new_data = np.empty(shape=(max_size,), dtype=obj.dtype)
                    new_data[: len(obj)] = obj
                    new_data[len(obj) :] = np.nan
                    data_dict[k] = new_data

        data_frame = pd.DataFrame(data_dict, index=data_index)
        dump_fn = functools.partial(self._dump_fn, data_frame)

        with pd.option_context(*self._pandas_display_options):
            perform_regression_check(
                datadir=self.datadir,
                original_datadir=self.original_datadir,
                request=self.request,
                check_fn=self._check_fn,
                dump_fn=dump_fn,
                extension=".csv",
                basename=basename,
                fullpath=fullpath,
                force_regen=self._force_regen,
            )



# encoding: UTF-8
import io
from functools import partial

import six

from pytest_regressions.common import perform_regression_check


class ImageRegressionFixture(object):
    """
    Regression test for image objects, accounting for small differences.
    """

    def __init__(self, datadir, original_datadir, request):
        """
        :type datadir: Path
        :type original_datadir: Path
        :type request: FixtureRequest
        """
        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self.force_regen = False

    def _load_image(self, filename):
        """
        Reads the image from the given file and convert it to RGB if necessary.
        This is necessary to be used with the ImageChops module operations.
        At this time, in this module, channel operations are only implemented
        for 8-bit images (e.g. "L" and "RGB").

        :param Path filename:
            The name of the file
        """
        from PIL import Image

        img = Image.open(six.text_type(filename), "r")
        if img.mode not in ("L" or "RGB"):
            return img.convert("RGB")
        else:
            return img

    def _compute_manhattan_distance(self, diff_image):
        """
        Computes a percentage of similarity of the difference image given.

        :param PIL.Image diff_image:
            An image in RGB mode computed from ImageChops.difference
        """
        import numpy

        number_of_pixels = diff_image.size[0] * diff_image.size[1]
        return (
            # To obtain a number in 0.0 -> 100.0
            100.0
            * (
                # Compute the sum of differences
                numpy.sum(diff_image)
                /
                # Divide by the number of channel differences RGB * Pixels
                float(3 * number_of_pixels)
            )
            # Normalize between 0.0 -> 1.0
            / 255.0
        )

    def _check_images_manhattan_distance(
        self, obtained_file, expected_file, expect_equal, diff_threshold
    ):
        """
        Compare two image by computing the differences spatially, pixel by pixel.

        The Manhattan Distance is used to compute how much two images differ.

        :param six.text_type obtained_file:
            The image with the obtained image

        :param six.text_type expected_files:
            The image with the expected image

        :param bool expected_equal:
            If True, the images are expected to be equal, otherwise, they're expected to be
            different.

        :param float diff_threshold:
            The maximum percentage of difference accepted.
            A value between 0.0 and 100.0

        :raises AssertionError:
            raised if they are actually different and expect_equal is False or
            if they are equal and expect_equal is True.
        """
        from PIL import ImageChops

        obtained_img = self._load_image(obtained_file)
        expected_img = self._load_image(expected_file)

        def check_result(equal, manhattan_distance):
            if equal != expect_equal:
                params = manhattan_distance, expected_file, obtained_file
                if expect_equal:
                    assert 0, (
                        "Difference between images too high: %.2f %%\n%s\n%s" % params
                    )
                else:
                    assert 0, (
                        "Difference between images too small: %.2f %%\n%s\n%s" % params
                    )

        # 1st check: identical
        diff_img = ImageChops.difference(obtained_img, expected_img)

        if diff_img.getbbox() is None:  # Equal
            check_result(True, None)

        manhattan_distance = self._compute_manhattan_distance(diff_img)
        equal = manhattan_distance <= diff_threshold
        check_result(equal, manhattan_distance)

    def check(self, image_data, diff_threshold=0.1, expect_equal=True, basename=None):
        """
        Checks that the given image contents are comparable with the ones stored in the data directory.

        :param bytes image_data: image data
        :param str|None basename: basename to store the information in the data directory. If none, use the name
            of the test function.
        :param bool expect_equal: if the image should considered equal below of the given threshold. If False, the
            image should be considered different at least above the threshold.
        :param float diff_threshold:
            Tolerage as a percentage (1 to 100) on how the images are allowed to differ.

        """
        from PIL import Image

        def dump_fn(target):
            image = Image.open(io.BytesIO(image_data))
            image.save(six.text_type(target), "PNG")

        perform_regression_check(
            datadir=self.datadir,
            original_datadir=self.original_datadir,
            request=self.request,
            check_fn=partial(
                self._check_images_manhattan_distance,
                diff_threshold=diff_threshold,
                expect_equal=expect_equal,
            ),
            dump_fn=dump_fn,
            extension=".png",
            basename=basename,
            force_regen=self.force_regen,
        )



def add_custom_yaml_representer(data_type, representer_fn):
    """
    Add custom representer to regression YAML dumper. It is polymorphic, so it works also for
    subclasses of `data_type`.

    :param type data_type: Type of objects.
    :param callable representer_fn: Function that receives object of `data_type` type as
        argument and must must return a YAML-convertible representation.
    """
    from .data_regression import RegressionYamlDumper

    return RegressionYamlDumper.add_custom_yaml_representer(data_type, representer_fn)



# encoding: UTF-8
from functools import partial

import six

from .common import perform_regression_check, check_text_files


class FileRegressionFixture(object):
    """
    Implementation of `file_regression` fixture.
    """

    def __init__(self, datadir, original_datadir, request):
        """
        :type datadir: Path
        :type original_datadir: Path
        :type request: FixtureRequest
        """
        self.request = request
        self.datadir = datadir
        self.original_datadir = original_datadir
        self.force_regen = False

    def check(
        self,
        contents,
        encoding=None,
        extension=".txt",
        newline=None,
        basename=None,
        fullpath=None,
        binary=False,
        obtained_filename=None,
        check_fn=None,
    ):
        """
        Checks the contents against a previously recorded version, or generate a new file.

        :param six.text_type contents: contents to write to the file
        :param six.text_type|None encoding: Encoding used to write file, if any.
        :param six.text_type extension: Extension of file.
        :param six.text_type|None newline: See `io.open` docs.
        :param bool binary: If the file is binary or text.
        :param obtained_filename: ..see:: FileRegressionCheck
        :param check_fn: a function with signature (obtained_filename, expected_filename) that should raise
            AssertionError if both files differ.
            If not given, use internal TODO.

        ..see: `data_regression.Check` for `basename` and `fullpath` arguments.
        """
        __tracebackhide__ = 0

        if binary and encoding:
            raise ValueError(
                "Only binary ({!r}) or encoding ({!r}) parameters must be passed at the same time.".format(
                    binary, encoding
                )
            )

        if binary:
            assert isinstance(
                contents, six.binary_type
            ), "Expected bytes contents but received type {}".format(
                type(contents).__name__
            )
        else:
            assert isinstance(
                contents, six.text_type
            ), "Expected text/unicode contents but received type {}".format(
                type(contents).__name__
            )

        import io

        if check_fn is None:

            if binary:

                def check_fn(obtained_filename, expected_filename):
                    if obtained_filename.read_bytes() != expected_filename.read_bytes():
                        raise AssertionError(
                            "Binary files {} and {} differ.".format(
                                obtained_filename, expected_filename
                            )
                        )

            else:
                check_fn = partial(check_text_files, encoding=encoding)

        def dump_fn(filename):
            mode = "wb" if binary else "w"
            with io.open(
                six.text_type(filename), mode, encoding=encoding, newline=newline
            ) as f:
                f.write(contents)

        perform_regression_check(
            datadir=self.datadir,
            original_datadir=self.original_datadir,
            request=self.request,
            check_fn=check_fn,
            dump_fn=dump_fn,
            extension=extension,
            basename=basename,
            fullpath=fullpath,
            force_regen=self.force_regen,
            obtained_filename=obtained_filename,
        )

    # non-PEP 8 alias used internally at ESSS
    Check = check



# encoding: UTF-8
from __future__ import unicode_literals
import textwrap
from pytest_regressions.common import Path

import pytest

from pytest_regressions.testing import check_regression_fixture_workflow


def test_simple_text_file(file_regression, datadir):
    contents = textwrap.dedent(
        """\
        # Title
        Another line: olá
    """
    )
    file_regression.check(contents, encoding="latin1", extension=".md")


def test_simple_bin_file(file_regression, datadir):
    contents = b"binary contents \xff\xff\xde"
    file_regression.check(contents, binary=True, extension=".bin")


def test_binary_and_text_error(file_regression):
    with pytest.raises(ValueError):
        file_regression.check("", encoding="UTF-8", binary=True)


def test_file_regression_workflow(testdir, monkeypatch):
    """
    :type testdir: _pytest.pytester.TmpTestdir
    :type monkeypatch: _pytest.monkeypatch.monkeypatch
    """
    import sys

    monkeypatch.setattr(sys, "get_data", lambda: "foo", raising=False)
    source = """
        import sys
        def test_1(file_regression):
            contents = sys.get_data()
            file_regression.check(contents, extension='.test')
    """

    def get_file_contents():
        fn = Path(str(testdir.tmpdir)) / "test_file" / "test_1.test"
        assert fn.is_file()
        return fn.read_text()

    check_regression_fixture_workflow(
        testdir,
        source,
        data_getter=get_file_contents,
        data_modifier=lambda: monkeypatch.setattr(
            sys, "get_data", lambda: "foobar", raising=False
        ),
        expected_data_1="foo",
        expected_data_2="foobar",
    )
